/**
 *
 */
package ca.uqam.info.latece.sboa.impl.algorithms;

import java.util.Hashtable;
import java.util.ArrayList;

import no.uib.cipr.matrix.Matrix;
import no.uib.cipr.matrix.Vector;
import no.uib.cipr.matrix.DenseVector;

import java.lang.management.MemoryMXBean;
import java.lang.management.ManagementFactory;

import org.semanticweb.owl.model.OWLEntity;
import org.semanticweb.owl.model.OWLOntology;

import ca.uqam.info.latece.sboa.impl.utils.OLATools;

import org.semanticweb.owl.align.Alignment;
import org.semanticweb.owl.align.AlignmentProcess;
import org.semanticweb.owl.align.Parameters;
import org.semanticweb.owl.align.AlignmentException;

import fr.inrialpes.exmo.align.impl.ObjectCell;
import fr.inrialpes.exmo.align.impl.BasicRelation;
import fr.inrialpes.exmo.align.impl.DistanceAlignment;
import fr.inrialpes.exmo.align.impl.method.StringDistances;

import ca.uqam.info.latece.sboa.impl.mgraph.MatchGraphsImpl;

import ca.uqam.info.latece.sboa.impl.weights.OLAWeightCapture;

import ca.uqam.info.latece.sboa.inter.ograph.Vertice;
import ca.uqam.info.latece.sboa.inter.mgraph.MatchGraphs;
import ca.uqam.info.latece.sboa.inter.ograph.VerticesPool;
import ca.uqam.info.latece.sboa.inter.ograph.OntologyGraph;
import ca.uqam.info.latece.sboa.inter.amatrix.AdjacencyMatrix;

import ca.uqam.info.latece.sboa.impl.amatrix.OLAMatrixUpdateImpl;

import ca.uqam.info.latece.sboa.inter.amatrix.OLAMatrixUpdate;

import ca.uqam.info.latece.sboa.impl.encodings.OLAOntologyGraphImpl;
import ca.uqam.info.latece.sboa.inter.encodings.OLAOntologyGraph;

import ca.uqam.info.latece.sboa.inter.algorithms.StructureBasedAlignment;

import ca.uqam.info.latece.sboa.inter.selection.AlignPairs;
import ca.uqam.info.latece.sboa.impl.selection.AlignPairsImpl;

/**
 * @author J.F. Djoufak
 *
 */
public class OLAlignment extends DistanceAlignment implements StructureBasedAlignment, AlignmentProcess {

	private Parameters par = null;
	private OntologyGraph oGraph1 = null;
	private OntologyGraph oGraph2 = null;
	private OntologyGraph mGraph = null;

	private String[] rootLab = {"Property","PropertyInstance","Relation"};
	private String[] rEdgeLab = {"subsumption","restriction"};
	private String[] remVer = {"Class", "Object", "DataType", "DataValue", "Token", "Cardinality"};
	private String[] remEdg = {"subsumption", "instantiation", "attribution", "domain", "range",
			"valuation", "all", "card", "name"};

	private Hashtable<Integer,Integer> leftCorr = null;
	private Hashtable<Integer,Integer> rightCorr = null;

	private VerticesPool leftPool = null;
	private VerticesPool rightPool = null;

	private OLAOntologyGraph oog = null;
	private OLAWeightCapture owg = null;
	private OLAMatrixUpdate omu = null;
	private OLATools ot = null;

	private long time = 0;
	private long newtime = 0;

	private boolean defaultInit = false;

	private String methodName = "levenshteinDistance";
	private String distanceClass = "fr.inrialpes.exmo.align.impl.method.StringDistances";

	private Hashtable<ArrayList<OWLEntity>,Double> alignRes = new Hashtable<ArrayList<OWLEntity>,Double>();

	public OLAlignment() {
		super();
	}

	public void align(Alignment align, Parameters param)
			throws AlignmentException {
		time = System.currentTimeMillis();
		par = param;
		defaultInit = false;

		owg = new OLAWeightCapture(par);
		oog = new OLAOntologyGraphImpl(par);
		ot = new OLATools(par);

		buildOGraph((OWLOntology)onto1.getOntology(),(OWLOntology)onto2.getOntology());
		buildMGraph(distanceClass,methodName);
		/*try{
			((OWLOntology)onto1.getOntology()).getOWLConnection().notifyOntologyDeleted((OWLOntology)onto1.getOntology());
			((OWLOntology)onto2.getOntology()).getOWLConnection().notifyOntologyDeleted((OWLOntology)onto2.getOntology());
			onto1 =null;
			onto2=null;
			}
		catch(Exception ex){
			ex.printStackTrace();
		}*/
		AdjacencyMatrix mMatrix = mGraph.getCategoryMatrix();
		int[] verIds = mMatrix.getLineVerticeIds();
		int sz= mMatrix.getNumberOfColums();
		Vector initVector = mGraph.getInitialVector(verIds,sz);
		leftPool = (VerticesPool)oGraph1.getVerticePool();
		rightPool = (VerticesPool)oGraph2.getVerticePool();
		prepareToCompute();
		Vector cbResult = computeCBSimilarity(mGraph,mMatrix,initVector);
		// reload
		extractResults(mGraph,cbResult,verIds,sz);
		clearMemory();
	}

	public void buildOGraph(OWLOntology onto1, OWLOntology onto2){
		oGraph1 = oog.build(onto1);
		oGraph2 = oog.build(onto2);
		String repS1 = oGraph1.concactLabels("Class")+oGraph1.concactLabels("Relation");
		String repS2 = oGraph2.concactLabels("Class")+oGraph2.concactLabels("Relation");
		double dist = StringDistances.levenshteinDistance(repS1,repS2);
		double sim = (1-dist);
		if (sim <= 0.17){
			defaultInit = true;
			String[] cRep = {"a","b","c","d","e","f","g","h","i"};
			String[] lRep = {"j","k","l","m","n","o","p","q","r","s"};
			oGraph1.setRepresentatives(cRep, lRep);
			oGraph2.setRepresentatives(cRep, lRep);
			oGraph1.buildAllDefaultNames(rootLab,rEdgeLab,remVer,remEdg);
			//System.err.println("done for onto 1");
			oGraph2.buildAllDefaultNames(rootLab,rEdgeLab,remVer,remEdg);
			//System.err.println("done for onto 2");
			updateOGraphs(false);
			methodName = "equalDistance";
		}
		else{
			updateOGraphs(true);
		}
		//oGraph1.createGraphMLFile("Ontology1", "/Users/djoufakj/Download/a1.graphml");
		//oGraph2.createGraphMLFile("Ontology1", "/Users/djoufakj/Download/a2.graphml");
	}

	public void buildMGraph(String distanceClass, String methodName) {
		String[] labs = oGraph1.getEdgePool().getPartsNames();
		String[] cats = oGraph1.getVerticePool().getPartsNames();
		MatchGraphs mg = new MatchGraphsImpl(oGraph1,oGraph2);
		mg.categoricalProduct(labs,cats,distanceClass,methodName,defaultInit);
		mGraph = mg.getOntologyGraph();
		leftCorr = mg.getLeftCorrespondences();
		rightCorr = mg.getRightCorrespondences();
		mg.clearAll();
		//mGraph.createGraphMLFile("MatchGraph", "/Users/djoufakj/Download/a12.graphml");
	}

	public void prepareToCompute(){
		clearOntologyGraphs();
		System.gc ();
	}

	public void viewMemoryUsage(String mess){
		newtime = System.currentTimeMillis();
		MemoryMXBean memoryBean = ManagementFactory.getMemoryMXBean();
		System.err.println(mess+": "+memoryBean.getHeapMemoryUsage());
		System.err.println("\t>>>> ELAPSED TIME = "+(newtime - time)+" <<<<");
		time = System.currentTimeMillis();
	}

	public Vector computeCBSimilarity(OntologyGraph mGraph, AdjacencyMatrix mMatrix,
			Vector initVect) {
		String[] cat = {"DataType", "DataValue", "Token", "Cardinality"};
		int[] staticVer = mGraph.getStaticWeightVertices(cat, 4);
		int staticSz = mGraph.getNumberOfStaticWeightVertices();
		int[] realId = mMatrix.getStaticLineVerticeIds(staticVer, staticSz);
		int[] verIds = mMatrix.getLineVerticeIds();
		int sz = mMatrix.getNumberOfLines();
		Matrix M = mMatrix.getMatrix();
		M = setMatrixEntries(M,realId,staticSz);
		mMatrix.setMatrix(M);
		omu = new OLAMatrixUpdateImpl(verIds,realId,M.numColumns(),leftCorr,rightCorr,
				leftPool,rightPool);
		boolean nextIter = true;
		Vector result = new DenseVector(sz);
		while (nextIter == true){
			M = omu.matrixUpdate(mGraph,mMatrix,initVect,true,true);
			result = M.mult(initVect,result);
			nextIter = convergenceNotObtained(result,initVect,sz);
			initVect = result.copy();
		}
		return result;
	}

	public boolean convergenceNotObtained(Vector newVect, Vector oldVect, int sz){
		boolean result = false;
		double epsilon = owg.getWeight(1);
		for(int i = 0 ; i < sz ; i++){
			double diff = newVect.get(i)-oldVect.get(i);
			if (diff>epsilon) return true;
		}
		return result;
	}

	public Alignment extractResults(OntologyGraph mGraph, Vector simVect, int[] indVect, int sz) {
		AlignPairs aps = new AlignPairsImpl();
		VerticesPool vPool = (VerticesPool)mGraph.getVerticePool();
		BasicRelation eq = new BasicRelation("=");
		//double threshold = owg.getWeight(0);
		alignRes.clear();
		for (int i = 0 ; i < sz ; i++){
			Vertice ver = (Vertice)vPool.get(indVect[i]);
			OWLEntity ent1 = ver.getFirstEntity();
			OWLEntity ent2 = ver.getSecondEntity();
			double val = simVect.get(i);
			boolean auth1 = ot.haveAuthorizedURI(ent1,true);
			boolean auth2 = ot.haveAuthorizedURI(ent2,true);
			boolean noInst = ver.getCategory().equalsIgnoreCase("Object")
								|| ver.getCategory().equalsIgnoreCase("PropertyInstance")
								|| ver.getCategory().equalsIgnoreCase("DataType")
								|| ver.getCategory().equalsIgnoreCase("DataValue")
								|| ver.getCategory().equalsIgnoreCase("Token")
								|| ver.getCategory().equalsIgnoreCase("Cardinality");
			if ((auth1 == true) && (auth2 == true) && (noInst == false)){
				aps.addPair(ent1,ent2,val);
			}
		}
		alignRes = aps.hungarianSelect(false);
		Object[] keys = alignRes.keySet().toArray();
		int ksz = keys.length;
		for (int j = 0 ; j < ksz ; j++){
		    ArrayList cple = (ArrayList)keys[j];
		    Object ent1 = (OWLEntity)cple.get(0);
		    Object ent2 = (OWLEntity)cple.get(1);
		    Double val = alignRes.get(keys[j]);
		    try {
			addCell(new ObjectCell((String)null,ent1,ent2,eq,val.doubleValue()));
		    } catch (AlignmentException ex) {
			ex.printStackTrace();
		    }
		}
		return ((Alignment) this);
	}

	public void buildLabelMatrices(OntologyGraph og1,
			OntologyGraph og2) {
	}

	public Vector computeFBSimilarity(OntologyGraph mGraph, AdjacencyMatrix mMatrix,
			Vector initVect) {
		return null;
	}

	private void updateOGraphs(boolean useLab){
		OLAOntologyGraph og1 = new OLAOntologyGraphImpl(par);
		Object[] vSet1 = oGraph1.getVerticePool().getAll();
		oGraph1 = og1.completeOntologyGraph(oGraph1, vSet1,useLab);
		OLAOntologyGraph og2 = new OLAOntologyGraphImpl(par);
		Object[] vSet2 = oGraph2.getVerticePool().getAll();
		oGraph2 = og2.completeOntologyGraph(oGraph2, vSet2,useLab);
	}

	private Matrix setMatrixEntries(Matrix mat, int[] indVect, int indsz){
		for (int i = 0 ; i < indsz ; i++){
			int indice = indVect[i];
			mat.set(indice,indice,1.0);
		}
		return mat;
	}

	private void clearOntologyGraphs(){
		oGraph1.clearAll();
		oGraph1 = null;
		oGraph2.clearAll();
		oGraph2 = null;
	}

	private void clearMemory(){
		par = null;oGraph1 = null;
		oGraph2 = null;mGraph = null;
		//rootLab = null;rEdgeLab = null;
		//remVer = null;remEdg = null;
		leftCorr = null; rightCorr = null;
		leftPool = null;rightPool = null;
		oog = null;owg = null;
		omu = null;ot = null;
		time = 0;newtime = 0;
		defaultInit = false;
		//methodName = null;distanceClass = null;
	}

}
