/**
 * 
 */
package ca.uqam.info.latece.sboa.impl.amatrix;

import java.util.TreeSet;
import java.util.Iterator;
import java.util.Hashtable;

import no.uib.cipr.matrix.Matrix;
import no.uib.cipr.matrix.sparse.FlexCompRowMatrix;

import ca.uqam.info.latece.sboa.inter.ograph.Edge;
import ca.uqam.info.latece.sboa.inter.ograph.Vertice;
import ca.uqam.info.latece.sboa.inter.ograph.VerticesPool;
import ca.uqam.info.latece.sboa.inter.amatrix.AdjacencyMatrix;


/**
 * @author J.F. Djoufak
 *
 */
public class AdjacencyMatrixImpl implements AdjacencyMatrix {
	
	private String name = null;
	private String type = null;
	
	private Matrix matrix = null;
	private int[] lineVerticeIds = null;
	private int[] colVerticeIds = null;
	private int[] nzLines = null;
	private int[] nzColumns = null;
	
	private int lines = 0;
	private int columns = 0;
	
	private Hashtable<Integer,Integer> lineIdToIndice = new Hashtable<Integer,Integer>();
	private Hashtable<Integer,Integer> colIdToIndice = new Hashtable<Integer,Integer>();

	public AdjacencyMatrixImpl() {
		
	}
	
	public AdjacencyMatrixImpl(Edge[] eSet){
		build(eSet);
	}
	
	public void build(Edge[] eSet) {
		insertAllVertices(eSet);
		int nbEdges = eSet.length;
		for (int i = 0 ; i < nbEdges ; i++){
			Edge edg = eSet[i];
			Integer Hd = new Integer(edg.getHead());
			int col = colIdToIndice.get(Hd).intValue();
			Integer Tl = new Integer (edg.getTail());
			int li = lineIdToIndice.get(Tl).intValue();
			double wgt = edg.getWeight();
			matrix.set(li,col,wgt);
		}
	}
	
	public void insertAllVertices(Edge[] eSet){
		TreeSet<Integer> foundLineIds = new TreeSet<Integer>();
		TreeSet<Integer> foundColIds = new TreeSet<Integer>();
		int nbEdges = eSet.length;
		lines = 0;
		columns = 0;
		for (int i = 0 ; i < nbEdges ; i++){
			Edge edg = eSet[i];
			Integer Hd = new Integer(edg.getHead());
			boolean resl = foundColIds.add(Hd);
			if (resl == true){columns++;}
			Integer Tl = new Integer(edg.getTail());
			boolean resc = foundLineIds.add(Tl);
			if (resc == true){lines++;}
		}
		lineVerticeIds = new int[lines];
		nzLines = new int[lines];
		for (int i = 0 ; i < lines ; i++){nzLines[i] = 0;}
		colVerticeIds = new int[columns];
		nzColumns = new int[columns];
		for (int i = 0 ; i < columns ; i++){nzColumns[i] = 0;}
		matrix = new FlexCompRowMatrix(lines,columns);
		insertVertice(foundLineIds, foundColIds);
	}
	
	public void insertVertice(TreeSet<Integer> foundLineIds,TreeSet<Integer> foundColIds){
		Iterator iterl = foundLineIds.iterator();
		int i = 0;
		while (iterl.hasNext()){
			Integer IdOfVertice = ((Integer)iterl.next());
			lineVerticeIds[i] = IdOfVertice.intValue();
			Integer LineOfMatrix = new Integer(i);
			lineIdToIndice.put(IdOfVertice,LineOfMatrix);
			i++;
		}
		Iterator iterc = foundColIds.iterator();
		i = 0;
		while (iterc.hasNext()){
			Integer IdOfVertice = ((Integer)iterc.next());
			colVerticeIds[i] = IdOfVertice.intValue();
			Integer LineOfMatrix = new Integer(i);
			colIdToIndice.put(IdOfVertice,LineOfMatrix);
			i++;
		}
	}
	
	public AdjacencyMatrixImpl(String nam, String typ, Matrix mat, int[] lidVect, int[] cidVect){
		build(nam,typ,mat,lidVect,cidVect);
	}
	
	public void build(String nam, String typ, Matrix mat, int[] lidVect, int[] cidVect){
		name = nam;
		type = typ;
		matrix = mat;
		lineVerticeIds = lidVect;
		colVerticeIds = cidVect;
		lines = matrix.numRows();
		columns = matrix.numColumns();
		for(int i = 0 ; i< lines ; i++){
			Integer Id = new Integer(lineVerticeIds[i]);
			Integer I = new Integer(i);
			lineIdToIndice.put(Id, I);
		}
		for(int i = 0 ; i< columns ; i++){
			Integer Id = new Integer(colVerticeIds[i]);
			Integer I = new Integer(i);
			colIdToIndice.put(Id, I);
		}
	}
	
	public int[] getLineVerticeIds(){
		return lineVerticeIds;
	}
	
	public int[] getColVerticeIds(){
		return colVerticeIds;
	}
	
	public int getLineIndice(int Id){
		return lineIdToIndice.get(new Integer(Id)).intValue();
	}
	
	public int getColumnIndice(int Id){
		return colIdToIndice.get(new Integer(Id)).intValue();
	}
	
	public void updateWithStaticLineVertice(int[] initialId, int sz){
		int[] realId = getStaticLineVerticeIds(initialId,sz);
		for (int i = 0 ; i < sz ; i++){
			int id = realId[i];
			matrix.set(id,id,1.0);
		}
	}
	
	public void updateWithStaticColVertice(int[] initialId, int sz){
		int[] realId = getStaticColVerticeIds(initialId,sz);
		for (int i = 0 ; i < sz ; i++){
			int id = realId[i];
			matrix.set(id,id,1.0);
		}
	}
	
	public int[] getStaticLineVerticeIds(int[] initialId, int sz){
		int[] result = new int[sz];
		for (int i= 0 ; i < sz ; i++){
			Integer I = new Integer(initialId[i]);
			result[i] = lineIdToIndice.get(I.intValue());
		}
		return result;
	}
	
	public int[] getStaticColVerticeIds(int[] initialId, int sz){
		int[] result = new int[sz];	
		for (int i= 0 ; i < sz ; i++){
			Integer I = new Integer(initialId[i]);
			result[i] = colIdToIndice.get(I.intValue());
		}
		return result;
	}

	public Matrix getMatrix() {
		return matrix;
	}

	public String getName() {
		return name;
	}
	
	public int getNumberOfColums(){
		return columns;
	}
	
	public int getNumberOfLines(){
		return lines;
	}

	public String getType() {
		return type;
	}

	public void setMatrix(Matrix mat) {
		matrix = mat;
	}

	public void setName(String nam) {
		name = nam;
	}
	
	public void setNumberOfColums(int col){
		columns = col;
	}
	
	public void setNumberOfLines(int lin){
		lines = lin;
	}

	public void setType(String typ) {
		type = typ;
	}
	
	public void viewMatrix(VerticesPool vPool){
		for (int i = 0 ; i < lines ; i++){
			for (int j = 0 ; j < columns ; j++){
				if (matrix.get(i,j)!=0){
					String Hd = ((Vertice)vPool.get(lineVerticeIds[i])).getLabel();
					String Tl = ((Vertice)vPool.get(colVerticeIds[j])).getLabel();
					System.err.println(Hd+"["+lineVerticeIds[i]+"]"+"<-"+Tl+"["+colVerticeIds[j]+"]"+"= "+matrix.get(i,j));
				}
			}
		}
	}
	
	public void clearAll(){
		name = null;
		type = null;
		matrix = null;
		lineVerticeIds = null;
		colVerticeIds = null;
		lineIdToIndice.clear();
		colIdToIndice.clear();
	}

}
