/**
 * 
 */
package ca.uqam.info.latece.sboa.impl.mgraph;

import java.util.ArrayList;
import java.util.Hashtable;

import org.semanticweb.owl.model.OWLEntity;

import ca.uqam.info.latece.sboa.inter.mgraph.MatchGraphs;

import ca.uqam.info.latece.sboa.inter.ograph.Edge;
import ca.uqam.info.latece.sboa.inter.ograph.Vertice;
import ca.uqam.info.latece.sboa.inter.ograph.EdgesPool;
import ca.uqam.info.latece.sboa.inter.ograph.VerticesPool;
import ca.uqam.info.latece.sboa.inter.ograph.OntologyGraph;

import ca.uqam.info.latece.sboa.impl.ograph.EdgeImpl;
import ca.uqam.info.latece.sboa.impl.ograph.VerticeImpl;
import ca.uqam.info.latece.sboa.impl.ograph.OntologyGraphImpl;

/**
 * @author J.F. Djoufak
 *
 */
public class MatchGraphsImpl implements MatchGraphs {

	private EdgesPool ePool1 = null;
	private EdgesPool ePool2 = null;
	private VerticesPool vPool1 = null;
	private VerticesPool vPool2 = null;
	
	private String[] lab1 = null;
	private String[] cat1 = null;
	private String[] lab2 = null;
	private String[] cat2 = null;
	
	private int lsz1 = 0;
	private int csz1 = 0;
	private int lsz2 = 0;
	private int csz2 = 0;
	
	private int vId = 0;
	private int eId = 0;
	private int hd = 0;
	private int tl = 0;
	
	private ArrayList<Vertice> vSet = new ArrayList<Vertice>();
	private ArrayList<Edge> eSet = new ArrayList<Edge>();
	
	private Hashtable<String,Integer> stringToId = new Hashtable<String, Integer>();
	private Hashtable<String, Integer> lab1ToId = new Hashtable<String, Integer>();
	private Hashtable<String, Integer> lab2ToId = new Hashtable<String, Integer>();
	private Hashtable<String, Integer> cat1ToId = new Hashtable<String, Integer>();
	private Hashtable<String, Integer> cat2ToId = new Hashtable<String, Integer>();
	
	private Hashtable<Integer,Integer> leftCorr = new Hashtable<Integer,Integer>();
	private Hashtable<Integer,Integer> rightCorr = new Hashtable<Integer,Integer>();
	
	private boolean defaultInit;
	
	public MatchGraphsImpl() {
		
	}
	
	public MatchGraphsImpl(OntologyGraph og1, OntologyGraph og2){
		init(og1,og2);
	}
	
	public void init(OntologyGraph og1, OntologyGraph og2){
		lsz1 = og1.getEdgePool().numberOfPart();
		lsz2 = og2.getEdgePool().numberOfPart();
		ePool1 = (EdgesPool)og1.getEdgePool();
		ePool2 = (EdgesPool)og2.getEdgePool();
		vPool1 = (VerticesPool)og1.getVerticePool();
		vPool2 = (VerticesPool)og2.getVerticePool();
		lab1 = ePool1.getPartsNames();
		lab2 = ePool2.getPartsNames();
		cat1 = vPool1.getPartsNames();
		cat2 = vPool2.getPartsNames();
		csz1 = vPool1.numberOfPart();
		csz2 = vPool2.numberOfPart();	
	}
	
	public void verticesProduct(String[] cats, String distanceClass, String method, boolean dInit){
		defaultInit = dInit;
		int csz = cats.length;
		boolean cTest = verticesCompatibilityTest(cats,csz,true);
		if (cTest == true){
			for(int i = 0 ; i < csz ; i++){
				int i1 = cat1ToId.get(cats[i]).intValue();
				Object[] vSet1 = vPool1.getAll(cat1[i1]);
				int sz1 = vPool1.partSize(cat1[i1]);
				int i2 = cat2ToId.get(cats[i]);
				Object[] vSet2 = vPool2.getAll(cat2[i2]);
				int sz2 = vPool2.partSize(cat2[i2]);
				productOfSameCategoryVertice(vSet1,sz1,vSet2,sz2,distanceClass,method,false);
			}
		}
	}
	
	private void productOfSameCategoryVertice(Object[] vSet1, int sz1, Object[] vSet2, int sz2, 
			String distanceClass, String method, boolean rTest){
		for (int j = 0 ; j < sz1 ; j++){
			System.err.println("here");
			Vertice v1 = (Vertice)vSet1[j];
			for (int k = 0 ; k < sz2 ; k++){
				Vertice v2 = (Vertice)vSet2[k];
				if (rTest == false){
					createVertice(v1,v2,distanceClass,method);
				}
				else{
					int neighSz = v2.getNeighborhoodSize(true) + v2.getNeighborhoodSize(false);
					if (neighSz > 0){
						createVertice(v1,v2,distanceClass,method);
					}
				}
			}
		}
	}
	
	private boolean verticesCompatibilityTest(String[] cats, int vsz, boolean clear){
		boolean result = true;
		if (clear == true){
			vId = 0;
			eId = 0;
			vSet.clear();
			eSet.clear();
		}
		cat1ToId.clear();
		cat2ToId.clear();
		for (int i = 0 ; i < csz1 ; i++){cat1ToId.put(cat1[i],i);}
		for (int j = 0 ; j < csz2 ; j++){cat2ToId.put(cat2[j],j);}
		for (int k = 0 ; k < vsz ; k++){
			String label = cats[k];
			if (cat1ToId.containsKey(label) && cat2ToId.containsKey(label)){
				int id1 = cat1ToId.get(label);
				int id2 = cat2ToId.get(label);
				result = result && (cat1[id1].equalsIgnoreCase(cat2[id2]));
			}
			else{
				result = false;
			}
		}
		return result;
	}

	public void categoricalProduct(String[] labs, String[] cats, String distanceClass, String method, boolean dInit) {
		defaultInit = dInit;
		int lsz = labs.length;
		boolean cTest = edgesCompatibilityTest(labs,lsz,true);
		if (cTest == true){
			for (int i = 0 ; i < lsz ; i++){
				int i1 = lab1ToId.get(labs[i]).intValue();
				int i2 = lab2ToId.get(labs[i]).intValue();
				Object[] eSet1 = ePool1.getAll(lab1[i1]);
				Object[] eSet2 = ePool2.getAll(lab2[i2]);
				int sz1 = ePool1.partSize(lab1[i1]);
				int sz2 = ePool2.partSize(lab2[i2]);
				createVerticesAndEdges(eSet1,eSet2,sz1,sz2,distanceClass,method);
			}
		}
		else{
			System.err.println("Input graphs are not compatible");
		}
	}
	
	private boolean edgesCompatibilityTest(String[] labs, int lsz, boolean clear){
		boolean result = true;
		if (clear == true){
			vId = 0;
			eId = 0;
			vSet.clear();
			eSet.clear();
		}
		lab1ToId.clear();
		lab2ToId.clear();
		for (int i = 0 ; i < lsz1 ; i++){lab1ToId.put(lab1[i],i);}
		for (int j = 0 ; j < lsz2 ; j++){lab2ToId.put(lab2[j],j);}
		for (int k = 0 ; k < lsz ; k++){
			String label = labs[k];
			if (lab1ToId.containsKey(label) && lab2ToId.containsKey(label)){
				int id1 = lab1ToId.get(label);
				int id2 = lab2ToId.get(label);
				result = result && (lab1[id1].equalsIgnoreCase(lab2[id2]));
			}
			else{
				result = false;
			}
		}
		return result;
	}
	
	private void createVerticesAndEdges(Object[] eSet1, Object[] eSet2, int sz1, int sz2, 
			String distanceClass, String method){
		for (int j = 0 ; j < sz1 ; j++){
			Edge e1 = (Edge)eSet1[j];
			Vertice e1Hd = (Vertice)vPool1.get(e1.getHead());
			Vertice e1Tl = (Vertice)vPool1.get(e1.getTail());
			for (int k = 0 ; k < sz2 ; k++){
				Edge e2 = (Edge)eSet2[k];
				Vertice e2Hd = (Vertice)vPool2.get(e2.getHead());
				Vertice e2Tl = (Vertice)vPool2.get(e2.getTail());
				boolean hdTest = sameEndPointCategoryTest(e1Hd,e2Hd);
				boolean tlTest = sameEndPointCategoryTest(e1Tl,e2Tl);
				if (hdTest == true){hd = createVertice(e1Hd,e2Hd,distanceClass,method);}
				if (tlTest == true){tl = createVertice(e1Tl,e2Tl,distanceClass,method);}
				if ((hdTest == true) && (tlTest == true)){createEdge(e1,e2);}
			}
		}
	}
	
	private boolean sameEndPointCategoryTest(Vertice v1, Vertice v2){
		boolean result = true;
		result = v1.getCategory().equalsIgnoreCase(v2.getCategory());
		return result;
	}
	
	private int createVertice(Vertice v1, Vertice v2, String distanceClass, String method){
		int result = vId;
		String sId = getIdentificationString(v1,v2);
		if (!stringToId.containsKey(sId)){
			String vCat = v1.getCategory();
			String labv1 = v1.getLabel();
			String labv2 = v2.getLabel();
			String vLab = "["+labv1+","+labv2+"]";
			double wgt = computeVerticeWeight(distanceClass,method,labv1.toLowerCase(),labv2.toLowerCase());
			OWLEntity oe1 = v1.getFirstEntity();
			OWLEntity oe2 = v2.getFirstEntity();
			Vertice ver = new VerticeImpl(result,vCat,vLab,"","",wgt,oe1,oe2);
			vSet.add(ver);
			stringToId.put(sId,result);
			createCorrespondences(result,v1.getId(),v2.getId());
			vId++;
		}
		else{
			result = stringToId.get(sId).intValue();
		}
		return result;
	}
	
	private double computeVerticeWeight(String distanceClass, String methodName, String s1, String s2){
		double result = 0.0;
		try{
			Class comparison = Class.forName(distanceClass);
			Class[] parTypes = {String.class, String.class};
			java.lang.reflect.Method m = comparison.getMethod(methodName, parTypes);
			Object[] parValues = {s1,s2};
			result = 1.0-((Double)m.invoke((Object)comparison, parValues)).doubleValue();
		}
		catch(Exception ex){
			ex.printStackTrace();
		}
		return result;
	}
	
	private void createCorrespondences(int mgId, int lId, int rId){
		Integer M = new Integer(mgId);
		Integer L = new Integer(lId);
		Integer R = new Integer(rId);
		leftCorr.put(M,L);
		rightCorr.put(M,R);
	}
	
	private void createEdge(Edge e1, Edge e2){
		String lab = e1.getLabel();
		double wgt = e1.getWeight() * e2.getWeight();
		Edge edg = new EdgeImpl (eId,lab,wgt,hd,tl);
		eSet.add(edg);
		eId++;
	}
	
	public Hashtable<Integer,Integer> getLeftCorrespondences(){
		return leftCorr;
	}
	
	public Hashtable<Integer,Integer> getRightCorrespondences(){
		return rightCorr;
	}
	
	public void clearAll(){
		ePool1.clearAll();
		ePool1 = null;
		ePool2.clearAll();
		ePool2 = null;
		lab1 = null;
		cat1 = null;
		lab2 = null;
		cat2 = null;
		vSet.clear();
		eSet.clear();
		stringToId.clear();
		lab1ToId.clear();
		lab2ToId.clear();
		cat1ToId.clear();
		cat2ToId.clear();
	}
	
	public OntologyGraph getOntologyGraph(){
		OntologyGraph result = new OntologyGraphImpl(cat1,lab1);
		
		Object[] ver = vSet.toArray();
		Object[] edg = eSet.toArray();
		result.StraightforwardBuild(ver,edg);
		
		return result;
	}
	
	private String getIdentificationString(Vertice v1, Vertice v2){
		String result = Integer.toString(v1.getId())+
						"x"+Integer.toString(v2.getId());
		return result;
	}

}
