/**
 * 
 */
package ca.uqam.info.latece.sboa.impl.ograph;

import java.util.TreeSet;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.ArrayList;

import java.io.FileWriter;
import java.io.PrintWriter;

import no.uib.cipr.matrix.Vector;
import no.uib.cipr.matrix.Matrix;
import no.uib.cipr.matrix.DenseVector;
import no.uib.cipr.matrix.sparse.FlexCompRowMatrix;

import ca.uqam.info.latece.sboa.inter.ograph.OntologyGraph;
import ca.uqam.info.latece.sboa.inter.ograph.Edge;
import ca.uqam.info.latece.sboa.inter.ograph.Vertice;
import ca.uqam.info.latece.sboa.inter.ograph.VerticesPool;
import ca.uqam.info.latece.sboa.inter.ograph.EdgesPool;
import ca.uqam.info.latece.sboa.inter.amatrix.AdjacencyMatrix;

import ca.uqam.info.latece.sboa.impl.ograph.VerticesPoolImpl;
import ca.uqam.info.latece.sboa.impl.ograph.EdgesPoolImpl;
import ca.uqam.info.latece.sboa.impl.amatrix.AdjacencyMatrixImpl;

/**
 * @author J.F. Djoufak
 *
 */
public class OntologyGraphImpl implements OntologyGraph {

	private VerticesPool vPool = null;
	private EdgesPool ePool = null;
	private int nOfStaticWeightVertices = 0;
	String[] catRep = null;
	String[] labRep = null;
	boolean annUsed = false;
	
	public OntologyGraphImpl() {
		
	}
	
	public OntologyGraphImpl(String[] categories, String[] labels){
		init(categories,labels);
	}
	
	public void init(String[] categories, String[] labels){
		int nbOfLab = labels.length;
		vPool = new VerticesPoolImpl(categories,nbOfLab);
		ePool = new EdgesPoolImpl(labels);
	}

	public void ImprovedBuild(Object[] vSet, Object[] eSet, String thesaurus) {
		
	}

	public void StraightforwardBuild(Object[] vSet, Object[] eSet) {
		vPool.insertAll(vSet);
		ePool.insertAll(eSet);
		upgradeVerticePool();
	}
	
	public EdgesPool getEdgePool(){
		return ePool;
	}
	
	public VerticesPool getVerticePool(){
		return vPool;
	}
	
	public void upgradeVerticePool(){
		String[] labNames = ePool.getPartsNames();
		int sz = ePool.numberOfPart();
		for (int i = 0 ; i < sz ; i++){
			String name = labNames[i];
			Object[] edgeVect = ePool.getAll(name);
			int vsz = ePool.partSize(name);
			for (int j = 0 ; j < vsz ; j++){
				Edge edg = (Edge)edgeVect[j];
				int eId = edg.getId();
				int hd = edg.getHead();
				int tl = edg.getTail();
				((Vertice)vPool.get(hd)).insertNeighId(tl,i,true);
				((Vertice)vPool.get(tl)).insertNeighId(hd,i,false);
				((Vertice)vPool.get(hd)).insertEdgeId(eId,i,true);
				((Vertice)vPool.get(tl)).insertEdgeId(eId,i,false);
			}
		}
	}

	public AdjacencyMatrix getCategoryAdjacencyMatrix(String cat1, String cat2) {
		Object[] candEdges = getLabelCandidate(cat1,cat2);
		HashSet<Integer> tailCand = getTailCandidate(cat1,cat2);
		ArrayList<Edge> selEdges = new ArrayList<Edge>();
		int csz = candEdges.length;
		for (int i = 0 ; i < csz ; i++){
			int eId = ((Integer)candEdges[i]).intValue();
			Edge cedg = (Edge)ePool.get(eId);
			int tail = cedg.getTail();
			if (tailCand.contains(new Integer(tail)) ){
				selEdges.add(cedg);
			}
		}
		int sz = selEdges.size();
		Edge[] edg = new Edge[sz];
		for (int k = 0 ; k<sz ; k++){edg[k] = (Edge)selEdges.get(k);}
		AdjacencyMatrix result = new AdjacencyMatrixImpl(edg);
		return result;
	}
	
	private Object[] getLabelCandidate(String cat1, String cat2){
		HashSet<Integer> result = new HashSet<Integer>();
		Object[] verSet1 = vPool.getAll(cat1);
		int vsz1 = vPool.partSize(cat1);
		for (int i = 0 ; i < vsz1 ; i++){
			ArrayList<Integer> eList = ((Vertice)verSet1[i]).getEdges(true); 
			result.addAll(eList);
		}
		return result.toArray();
	}
	
	private HashSet<Integer> getTailCandidate(String cat1, String cat2){
		HashSet<Integer> result = new HashSet<Integer>();
		Object[] verSet2 = vPool.getAll(cat2);
		int vsz2 = vPool.partSize(cat2);
		for (int j = 0 ; j < vsz2 ; j++){
			Integer J = ((Vertice)verSet2[j]).getId();
			result.add(J);
		}
		return result;
	}
	

	public AdjacencyMatrix getCategoryMatrix() {
		String[] cat = vPool.getPartsNames();
		int nOfP = vPool.numberOfPart();
		Hashtable<Integer,Integer> IdToInd = new Hashtable<Integer,Integer>();
		int[] matrixIds = getCategoryMatrixVerticeIds(cat,nOfP,IdToInd);
		Matrix data = createCategoryMatrix(cat,nOfP,matrixIds.length,IdToInd);
		AdjacencyMatrix result = new AdjacencyMatrixImpl("CategoryMatrix","Matrix",data,
				matrixIds,matrixIds);
		return result;
	}
	
	public int[] getCategoryMatrixVerticeIds(String[] cat, int nbParts, 
			Hashtable<Integer,Integer> IdToIndices){
		TreeSet<Integer> ids = new TreeSet<Integer>();
		ArrayList<Integer> orId = new ArrayList<Integer>();
		for (int i = 0 ; i < nbParts ; i++){
			Object[] catVer = vPool.getAll(cat[i]);
			int catsz = vPool.partSize(cat[i]);
			ids.clear();
			for (int j = 0 ; j < catsz ; j++){
				Vertice ver = (Vertice)catVer[j];
				ids.add(ver.getId());
			}
			Iterator iter = ids.iterator();
			while (iter.hasNext()){
				Integer I = (Integer)iter.next();
				orId.add(I);
			}	
		}
		int sz = vPool.size();
		int [] result = new int[sz];
		for (int i = 0 ; i < sz ; i++){
			int Id = orId.get(i).intValue();
			result[i] = Id;
			IdToIndices.put(Id, new Integer(i));
		}
		return result;
	}
	
	public Matrix createCategoryMatrix(String[] cat, int nbParts, int nOfLines, 
			Hashtable<Integer,Integer> IdToIndices){
		FlexCompRowMatrix result = new FlexCompRowMatrix(nOfLines,nOfLines);
			Object[] edg = ePool.getAll();
			int sz = ePool.size();
			for (int j = 0 ; j < sz ; j++){
				Edge e = (Edge)edg[j];
				int hd = e.getHead();
				int tl = e.getTail();
				int rj = (IdToIndices.get(new Integer(hd))).intValue();
				int ri = (IdToIndices.get(new Integer(tl))).intValue();
				result.set(ri,rj,e.getWeight());
			}		
		return result;
	}

	public AdjacencyMatrix getLabelAdjacencyMatrix(String lab) {
		Object[] labEdgO = ePool.getAll(lab);
		int sz = labEdgO.length;Edge[] labEdg = new Edge[sz];
		for (int i = 0 ; i<sz ; i++){labEdg[i] = (Edge)labEdgO[i];}
		AdjacencyMatrix result = new AdjacencyMatrixImpl(labEdg);
		return result;
	}

	public AdjacencyMatrix[] getLabelMatrices() {
		int nOfP = ePool.numberOfPart();
		AdjacencyMatrix[] result = new AdjacencyMatrix[nOfP];
		String[] lab = ePool.getPartsNames();
		for (int i = 0 ; i < nOfP ; i++){
			result[i] = getLabelAdjacencyMatrix(lab[i]);
		}
		return result;
	}
	
	public Vector getInitialVector(int[] verIds, int sz){
		DenseVector result = new DenseVector(sz);
		for (int i = 0 ; i < sz ; i++){
			Vertice ver = (Vertice)vPool.get(verIds[i]);
			result.add(i, ver.getWeight());
		}
		return result;
	}
	
	public int[] getStaticWeightVertices(String[] statCat, int sz){
		nOfStaticWeightVertices = 0;
		for (int i = 0 ; i < sz ; i++){
			nOfStaticWeightVertices = nOfStaticWeightVertices + vPool.partSize(statCat[i]);
		}
		int [] result = new int[nOfStaticWeightVertices];
		int currIndex = 0;
		for (int j = 0 ; j < sz ; j++){
			ArrayList<Integer> vId = vPool.getAllIds(statCat[j]);
			int psz = vPool.partSize(statCat[j]);
			for (int k = 0 ; k < psz ; k++){
				Integer I = vId.get(k);
				result[currIndex] = I.intValue();
				currIndex++;
			}
		}
		return result;
	}
	
	public int getNumberOfStaticWeightVertices(){
		return nOfStaticWeightVertices;
	}

	public void insertEdge(Object edg) {
		ePool.insert(edg);
	}

	public void insertEdge(Object[] eSet) {
		ePool.insertAll(eSet);
	}

	public void insertVertice(Object ver) {
		vPool.insert(ver);
	}

	public void insertVertice(Object[] vSet) {
		vPool.insertAll(vSet);
	}
	
	public void getResume(){
		System.err.println("Number of vertices = "+vPool.size());
		int nbCat = vPool.numberOfPart();
		String[] catNames = vPool.getPartsNames();
		for (int i = 0 ; i < nbCat ; i++){
			String name = catNames[i];
			System.err.println("\t"+name+": "+vPool.partSize(name)+" vertices");
		}
		System.err.println("Number of edges = "+ePool.size());
		int nbLab = ePool.numberOfPart();
		String[] labNames = ePool.getPartsNames();
		for (int j = 0 ; j < nbLab ; j++){
			String name = labNames[j];
			System.err.println("\t"+name+": "+ePool.partSize(name)+" edges");
		}
	}
	
	public String concactLabels(String lab){
		HashSet<String> ht = new HashSet<String>();
		Object[] tvSet = vPool.getAll(lab);
		int tvSz = vPool.partSize(lab);
		for (int i = 0 ; i < tvSz ; i++){
			Vertice tokVer = (Vertice)tvSet[i];
			ht.add(tokVer.getLabel());
		}
		String result = ht.toString();
		result = result.replace(" ","");result = result.replace(",","");
		result = result.replace("[","");result = result.replace("]","");
		return result.toString();
	}
	
	private String getVerticeConfig(Vertice ver){
		String result = "";
		int sz = ePool.numberOfPart();
		for(int i = 0 ; i < sz ; i++){
			int psz = ver.getNeighborhoodSize(i,true);
			result = result+Integer.toString(psz);
		}
		result = result+":";
		for(int i = 0 ; i < sz ; i++){
			int psz = ver.getNeighborhoodSize(i,false);
			result = result+Integer.toString(psz);
		}
		return result;
	}
	
	public String defaultName(Vertice ver, String[] repr, int[] lIds, int rSz, TreeSet<Integer> reach){
		String result = ver.getComments();
		String config = getVerticeConfig(ver);
		int ngSz = 0;
		for (int i = 0 ; i < rSz ; i++){
			ngSz = ngSz + ver.getNeighborhoodSize(lIds[i],true);
		}
		if ((ngSz > 0)&&(result.equalsIgnoreCase(""))){
			for (int i = 0 ; i < rSz ; i++){
				ArrayList<Integer> hN = ver.getNeighborhood(lIds[i],true);
				int hNsz = ver.getNeighborhoodSize(lIds[i],true);
				for (int j = 0 ; j < hNsz ; j++){
					int nId = hN.get(j).intValue();
					Vertice nVer = (Vertice)vPool.get(nId);
					String cat = ver.getCategory();
					String rep = catRep[vPool.getPartId(cat)];
					//TreeSet<Integer> ts = new TreeSet<Integer>(nVer.getNeighborhood(lIds[i],true));
					Integer I = new Integer(nVer.getId());
					boolean test = reach.contains(I);//ts.contains(new Integer(ver.getId()));
					if ((ver.getId() != nVer.getId()) && (test == false)){
						reach.add(I);
						result = result + "("+defaultName(nVer,repr,lIds,rSz,reach)+"+"+repr[i]+":"+rep+config+")";//":"+j+
					}
				}
			}
		}
		else{
			if (ngSz == 0){
				String cat = ver.getCategory();
				String rep = catRep[vPool.getPartId(cat)];
				result = "(x:"+rep+config+")";
			}
		}
		ver.setComments(result);
		return result;
	}
	
	public void setRepresentatives(String[] cRep, String[] lRep){
		int csz = vPool.numberOfPart();
		int lsz = ePool.numberOfPart();
		if ((csz != cRep.length)||(lsz != lRep.length)){
			System.err.println("Strict representatives needed");
			System.exit(1);
		}
		else{
			catRep = new String[csz];
			labRep = new String[lsz];
			for (int i = 0 ; i < csz;i++){catRep[i] = cRep[i];}
			for(int j = 0 ; j < lsz ; j++){labRep[j] = lRep[j];}
		}
	}
	
	public String[] buildRepresentatives(String[] lab){
		int lSz = lab.length;
		String[] result = new String[lSz];
		for (int i = 0 ; i < lSz ; i++){
			int id = ePool.getPartId(lab[i]);
			result[i] = labRep[id];
		}
		return result;
	}
	
	private void buildDefaultNameOfParts(String[] rootLab, String[] rEdgeLab){
		int rootSz = 0;
		int rootNb = rootLab.length;
		int rEdgeNb = rEdgeLab.length;
		int[] lIds = new int[rEdgeNb];
		Object[] roots =  buildRoots(rootLab);
		String[] rootRep = buildRepresentatives(rEdgeLab);
		for (int i = 0 ; i < rootNb ; i++){rootSz = rootSz + vPool.partSize(rootLab[i]);}
		for (int i = 0 ; i < rEdgeNb ; i++){lIds[i] = ePool.getPartId(rEdgeLab[i]);}
		for (int i = 0 ; i < rootSz ; i++){
			Vertice ver = (Vertice)roots[i];
			TreeSet<Integer> reach = new TreeSet<Integer>();
			reach.add(new Integer (ver.getId()));
			defaultName(ver,rootRep,lIds,rEdgeNb,reach);
		}
	}
	
	public void buildAllDefaultNames(String[] rootLab, String[] rEdgeLab, String[] remVer, String[] remEdg){
		buildDefaultNameOfParts(rootLab,rEdgeLab);
		buildDefaultNameOfParts(remVer,remEdg);
	}
	
	public Object[] buildRoots(String[] rootLab){
		ArrayList<Vertice> result = new ArrayList<Vertice>();
		int rSz = rootLab.length;
		for (int i = 0 ; i < rSz ; i++){
			Object[] data = vPool.getAll(rootLab[i]);
			int dSz = vPool.partSize(rootLab[i]);
			for (int j = 0 ; j < dSz ; j++){result.add((Vertice)data[j]);}
		}
		return result.toArray();
	}
	
	public void createGraphMLFile(String name, String fpath){
		try{
			FileWriter fstream = new FileWriter(fpath);
			PrintWriter out = new PrintWriter(fstream);
			initGraphMLFile(out,name);
			insertVerticesInGraphMLFile(out);
			insertEdgesInGraphMLFile(out);
			finalizeGraphMLFile(out);
			out.close();
		}
		catch(Exception ex){
			ex.printStackTrace();
		}
	}
	
	public void clearAll(){
		ePool.clearAll();
		ePool = null;
	}
	
	public void initGraphMLFile(PrintWriter out, String name){
		out.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
		out.println("<graphml xmlns=\"http://graphml.graphdrawing.org/xmlns/graphml\" " +
				"xmlns:y=\"http://www.yworks.com/xml/graphml\" " +
				"xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " +
				"xsi:schemaLocation=\"http://graphml.graphdrawing.org/xmlns/graphml " +
				"http://www.yworks.com/xml/schema/graphml/1.0/ygraphml.xsd\">");
		out.println("<key id=\"d0\" for=\"node\" yfiles.type=\"nodegraphics\"/>");
		out.println("<key id=\"d1\" for=\"edge\" yfiles.type=\"edgegraphics\"/>");
		out.println("<graph id=\""+name+"\" edgedefault=\"directed\">");
	}
	
	public void insertVerticesInGraphMLFile(PrintWriter out){
		String [] categories = vPool.getPartsNames();
		int sz = categories.length;
		for (int i = 0 ; i < sz ; i++){
			Object[] cVer = vPool.getAll(categories[i]);
			int csz = cVer.length;
			for (int j = 0 ; j < csz ; j++){
				Vertice ver = (Vertice)cVer[j];
				int id = ver.getId();
				String name = ver.getLabel();
				out.println("<node id=\""+"n"+id+"\">");
					out.println("\t<data key=\"d0\" >");
						out.println("\t\t<y:ShapeNode >");
								out.println("\t\t\t<y:NodeLabel>"+name+"</y:NodeLabel>");
								out.println("<y:Shape type=\""+getVerticeShape(categories[i])+"\"/>");
						out.println("\t\t</y:ShapeNode>");
					out.println("\t</data>");
				out.println("</node>");
			}
		}
	}
	
	public void insertEdgesInGraphMLFile(PrintWriter out){
		String [] labels = ePool.getPartsNames();
		int sz = labels.length;
		for (int i = 0 ; i < sz ; i++){
			Object[] lEdg = ePool.getAll(labels[i]);
			int lsz = lEdg.length;
			for (int j = 0 ; j < lsz ; j++){
				Edge edg = (Edge)lEdg[j];
				int id = edg.getId();
				String lab = edg.getLabel();
				int hd = edg.getHead();
				int tl = edg.getTail();
				out.println("<edge id=\""+"e"+id+"\" source=\""+"n"+hd+"\" target=\""+"n"+tl+"\">");
					out.println("\t<data key=\"d1\" >");
						out.println("\t\t<y:PolyLineEdge >");
							out.println("\t\t\t<y:EdgeLabel>"+lab+"</y:EdgeLabel>");
							out.println("\t\t\t<y:BendStyle smoothed=\"false\"/>");
						out.println("\t\t</y:PolyLineEdge>");
					out.println("\t</data>");
				out.println("</edge>");
			}
		}
	}
	
	public String getVerticeShape(String category){
		String result=null;
		
		if (category.equalsIgnoreCase("Class")){result = "rectangle";}
		else{
			if (category.equalsIgnoreCase("Token")){result = "octagon";}
			else{
				if (category.equalsIgnoreCase("Relation")){result = "hexagon";}
				else{
					if (category.equalsIgnoreCase("Property")){result = "parallelogram";}
					else{
						if (category.equalsIgnoreCase("DataType")){result = "diamond";}
						else{
							if (category.equalsIgnoreCase("Cardinality")){result = "roundrectangle";}
							else{
								if (category.equalsIgnoreCase("Object")){result = "triangle";}
								else{
									if (category.equalsIgnoreCase("PropertyInstance")){result = "trapezoid";}
									else{
										result = "ellipse";
									}
								}
							}
						}
					}
				}
			}
		}
		
		return result;
	}
	
	public void finalizeGraphMLFile(PrintWriter out){
		out.println("</graph>");
		out.println("</graphml>");
	}
	
	public void viewAll(){
		vPool.viewAllVertices();
		ePool.viewAllEdges();
	}
	
}
