/**
 * 
 */
package ca.uqam.info.latece.sboa.impl.ograph;

import java.util.Vector;
import java.util.ArrayList;
import java.util.Hashtable;

import ca.uqam.info.latece.sboa.inter.ograph.VerticesPool;
import ca.uqam.info.latece.sboa.inter.ograph.Vertice;

/**
 * @author J.F. Djoufak
 *
 */
public class VerticesPoolImpl implements VerticesPool {
	
	private String[] categoryNames = null;
	
	private Hashtable<String,Integer> categoryLabelToId = new Hashtable<String,Integer>();
	private Hashtable<String,Integer> verticeLabelToId = new Hashtable<String,Integer>();
	private Vector<ArrayList<Integer>> IdsGroupedByCategory = null;
	private Hashtable<Integer,Vertice> verticeIdToVertice = new Hashtable<Integer,Vertice>();
	
	private int[] IdGroupsLen = null;
	private int size = 0;
	private int numberOfCategories = 0;
	private int numberOfNeighParts = 0;
	

	public VerticesPoolImpl(){
	}
	
	public VerticesPoolImpl(String[] cat, int nOfLab){
		init(cat);
		setNeighborhoodParts(nOfLab);
	}
	
	public int numberOfPart() {
		return numberOfCategories;
	}
	
	public int getPartId(String label){
		return categoryLabelToId.get(label).intValue();
	}
	
	public boolean containsCategory(String cat){
		return categoryLabelToId.containsKey(cat);
	}

	public boolean contains(String lab) {
		Integer Id = (Integer)verticeLabelToId.get(lab);
		return verticeIdToVertice.containsKey(Id);
	}

	public boolean contains(int Id) {
		Integer I = new Integer(Id);
		return verticeIdToVertice.containsKey(I);
	}

	public Object get(String lab) {
		Integer Id = (Integer)verticeLabelToId.get(lab);
		return verticeIdToVertice.get(Id);
	}

	public Object get(int Id) {
		Integer I = new Integer(Id);
		return verticeIdToVertice.get(I);
	}
	
	public ArrayList<Integer> getAllIds(String lab){
		Integer catIndice = (Integer)categoryLabelToId.get(lab);
		return (ArrayList<Integer>)IdsGroupedByCategory.elementAt(catIndice.intValue());
	}

	public Object[] getAll(String lab) {
		Integer catIndice = (Integer)categoryLabelToId.get(lab);
		int catLen = IdGroupsLen[catIndice.intValue()];
		Object[] result = new Object[catLen];
		ArrayList<Integer> idOfCat = (ArrayList<Integer>)IdsGroupedByCategory.elementAt(catIndice.intValue());
		for(int i=0 ; i < catLen ; i++){
			Integer I = (Integer)idOfCat.get(i);
			result[i] = get(I.intValue());
		}
		return result;
	}
	
	public Object[] getAll(){
		ArrayList<Object> vSet = new ArrayList<Object>();
		for (int i = 0 ; i < numberOfCategories ; i++){
			Object[] data = getAll(categoryNames[i]);
			int dSz = this.partSize(categoryNames[i]);
			for (int j = 0 ; j < dSz ; j++){
				vSet.add(data[j]);
			}
		}
		return vSet.toArray();
	}

	public Object[] getAll(int[] Id) {
		int nbIds = Id.length;
		Object[] result = new Object[nbIds];
		for(int i =0 ; i < nbIds ; i++){
			result[i] = get(Id[i]);
		}
		return result;
	}
	
	public String[] getPartsNames(){
		return categoryNames;
	}

	public void init(String[] cat) {
		int csize = cat.length;
		categoryNames = new String[csize];
		IdsGroupedByCategory = new Vector<ArrayList<Integer>>(csize);
		IdGroupsLen = new int[csize];
		for(int i = 0 ; i<csize ; i++) {
			categoryNames[i] = cat[i];
			categoryLabelToId.put(categoryNames[i],i);
			IdsGroupedByCategory.insertElementAt(new ArrayList<Integer>(),i);
			IdGroupsLen[i] = 0;
			numberOfCategories++;
		}
	}

	public void insert(Object obj) {
		Vertice ver = (Vertice) obj;
		String vlab = ver.getLabel();
		int vid = ver.getId();
		String vcat = ver.getCategory();
		ver.setNumberOfEdgesLabels(numberOfNeighParts);
		Integer catId = (Integer)categoryLabelToId.get(vcat);
		Integer Id = new Integer(vid);
		IdsGroupedByCategory.elementAt(catId.intValue()).add(Id);
		verticeLabelToId.put(vlab,Id);
		verticeIdToVertice.put(Id,ver);
		IdGroupsLen[catId.intValue()]++;
		size++;
	}

	public void insertAll(Object[] objSet) {
		int oSize = objSet.length;
		for(int i = 0 ; i < oSize ; i++){
			insert(objSet[i]);
		}
	}

	public boolean isEmpty() {
		return (size == 0);
	}
	
	public int partSize(String name){
		Integer Id = (Integer)categoryLabelToId.get(name);
		return IdGroupsLen[Id.intValue()];
	}

	public void remove(String lab) {
		Vertice ver = (Vertice)get(lab);
		String cat = ver.getCategory();
		Integer cId = (Integer)categoryLabelToId.get(cat);
		Integer vId = (Integer)verticeLabelToId.get(lab);
		IdsGroupedByCategory.elementAt(cId).remove(vId);
		IdGroupsLen[cId.intValue()]--;
		size--;
		verticeIdToVertice.remove(vId);
		verticeLabelToId.remove(lab);
	}

	public void remove(int Id) {
		Vertice ver = (Vertice)get(Id);
		String cat = ver.getCategory();
		Integer cId = (Integer)categoryLabelToId.get(cat);
		Integer I = new Integer(Id);
		IdsGroupedByCategory.elementAt(cId).remove(I);
		IdGroupsLen[cId.intValue()]--;
		size--;
		verticeIdToVertice.remove(I);
		String lab = ver.getLabel();
		verticeLabelToId.remove(lab);
	}

	public void removeAll(String lab) {
		Integer cId = (Integer)categoryLabelToId.get(lab);
		int cSize = IdGroupsLen[cId.intValue()];
		ArrayList<Integer> labList = (ArrayList<Integer>)IdsGroupedByCategory.elementAt(cId.intValue());
		for (int i = 0 ; i < cSize ; i++){
			Integer I = (Integer)labList.get(i);
			remove(I.intValue());
		}
	}

	public void removeAll(int[] Id) {
		int idSize = Id.length;
		for (int i = 0 ; i < idSize ; i++){
			remove(Id[i]);
		}
	}

	public int size() {
		return size;
	}
	
	public int getNumberOfNeighborhooParts(){
		return numberOfNeighParts;
	}
	
	public void setNeighborhoodParts(int nOfLab){
		numberOfNeighParts = nOfLab;
	}
	
	public void viewAllVertices(){
		for (int i = numberOfCategories-1 ; i >= 0 ; i--){
			String name = categoryNames[i];
			Object[] verVect = getAll(name);
			int sz = partSize(name);
			for (int j = 0 ; j < sz ; j++){
				((Vertice)verVect[j]).view();
			}
		}
	}
	
	public void clearAll(){
		categoryNames = null;
		categoryLabelToId.clear();
		verticeLabelToId.clear();
		IdsGroupedByCategory.clear();
		verticeIdToVertice.clear();
		IdGroupsLen = null;
	}

}
