package ca.uqam.info.latece.sboa.impl.utils;

import java.net.URI;

import java.util.Map;
import java.util.Set;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

import org.semanticweb.owl.model.OWLEntity;
import org.semanticweb.owl.model.OWLOntology;
import org.semanticweb.owl.align.Parameters;
import org.semanticweb.owl.impl.model.OWLOrImpl;
import org.semanticweb.owl.impl.model.OWLClassImpl;
import org.semanticweb.owl.impl.model.OWLIndividualImpl;
import org.semanticweb.owl.impl.model.OWLConcreteDataImpl;
import org.semanticweb.owl.impl.model.OWLDataPropertyImpl;
import org.semanticweb.owl.impl.model.OWLObjectPropertyImpl;
import org.semanticweb.owl.impl.model.OWLConcreteDataTypeImpl;

/**
 * @author J.F. Djoufak
 *
 */
public class OLATools {
	
	int nullVerticeUriId = 0;
	private String[] excludedURIs = null;

	public OLATools(Parameters param) {
	    String ex = (String) param.getParameter("excludeURIs");
	    if ( ex != null && !ex.equals("") ) excludedURIs = ex.split("::");
	}
	
	public String getName(URI uri, String type, int gen){
		String result = "";
		if(uri != null){
			result = uri.getFragment();
			if (result == null) {
				String path = uri.getPath();
				StringTokenizer st = new StringTokenizer(path,"/");
				while (st.hasMoreTokens()){
					result = st.nextToken();
				}
			}
		}
		else{
			if (gen == 1){
				result = type+String.valueOf(nullVerticeUriId);
				nullVerticeUriId++;
			}
		}
		return result;
	}
	
	public String getType(OWLIndividualImpl owlI, OWLOntology onto, int choice){
		String result = "";
		Set type = owlI.getTypes(onto);
		Object[] typeA = type.toArray();
		int sz = typeA.length;
		for (int i = 0 ; i < sz ; i++){
			OWLClassImpl oc = (OWLClassImpl)typeA[i];
			if (oc != null){
				if (choice == 0){
					result = result+oc.getURI().toString();
				}
				else{
					result = result+getName(oc.getURI(),"",1);
				}
			}
			
		}
		return result;
	}

	public String getIdentification(OWLIndividualImpl ind, OWLOntology onto){
		String result = "";
		if (ind.getURI()== null){
			Map dpmap = ind.getDataPropertyValues(onto);
			Object[] di = dpmap.keySet().toArray();
			for(int i=0;i<di.length;i++){
				OWLDataPropertyImpl dpi = (OWLDataPropertyImpl)di[i];
				HashSet val = (HashSet)dpmap.get(dpi);
				Iterator iter = val.iterator();
				while (iter.hasNext()){
					OWLConcreteDataImpl cdi = (OWLConcreteDataImpl)iter.next();
					result = result + cdi.getValue().toString();
				}
			}
			Map opmap = ind.getObjectPropertyValues(onto);
			Object[] dj = opmap.keySet().toArray();
			for(int i=0;i<dj.length;i++){
				OWLObjectPropertyImpl opi = (OWLObjectPropertyImpl)dj[i];
				HashSet val = (HashSet)opmap.get(opi);
				Iterator iter = val.iterator();
				while (iter.hasNext()){
					OWLIndividualImpl cdi = (OWLIndividualImpl)iter.next();
					result = result + getName(cdi.getURI(),"",0);
				}
			}
		}
		return result;
	}
	
	public Object[] getOperand (Object op){
		Object[] operand = null;
		if (op instanceof OWLOrImpl){
			OWLOrImpl oor = (OWLOrImpl) op;
			operand = oor.getOperands().toArray();
		}
		else{
			operand = new Object[1];
			operand[0] = op;
		}
		return operand;
	}
	
	public String generateCardinalityString(int least, int most){
		String card = "";
		if(least >= 0){
			card = card + Integer.toString(least);
		}
		else{
			card = card + Integer.toString(0);
		}
		if (most >= 0){
			card = card + ":"+ Integer.toString(most);
		}
		return card;
	}
	
	public String getDataPropertyRange(OWLDataPropertyImpl dpi, OWLOntology onto){
		String result = null;
		Object[] ran = dpi.getRanges(onto).toArray();
		URI dturi = null;
		if (ran.length>0){
			dturi= ((OWLConcreteDataTypeImpl)ran[0]).getURI();
		}
		if (dturi==null){
			result = "null";
		}
		else{
			result = getName(dturi,"",0);
		}
		return result;
	}
	
	public String getObjectPropertyRange(OWLObjectPropertyImpl dpi, OWLOntology onto){
		String result = null;
		Object[] ran = dpi.getRanges(onto).toArray();
		URI dturi = null;
		if (ran.length>0){
			dturi= ((OWLClassImpl)ran[0]).getURI();
		}
		if (dturi==null){
			result = "null";
		}
		else{
			result = getName(dturi,"",0);
		}
		return result;
	}
	
	public boolean stringFilter(String data){
		boolean result = false;
		if (data != null){
			int i1 = data.toLowerCase().indexOf("zimmermann");
			int i2 = data.toLowerCase().indexOf("euzenat");
			int i3 = data.toLowerCase().indexOf("knouf");
			int i4 = data.toUpperCase().indexOf(".rdf");
			if ( (i1 != -1) || (i2 != -1) || (i3 != -1) || (i4 != -1)){
				result = true;
			}
		}
		else{
			result = true;
		}
		
		return result;
	}
	
	public boolean haveAuthorizedURI(OWLEntity ent, boolean exc){
	    String uri = "";
	    if (ent == null) {return false;}
	    if (excludedURIs != null) {
	    	try { 
	    		URI ruri = ent.getURI();
	    		if ( ruri == null ) return false;
					uri = ruri.toString();
	    		}
	    	catch (Exception e) { 
	    		e.printStackTrace();
	    	}
	    	if ( uri.indexOf('#') >= 0 ){
	    		uri = uri.substring(0, uri.indexOf('#'));
	    		}
	    	for (int x = excludedURIs.length - 1; x >= 0; x--) {
	    		if (uri.startsWith(excludedURIs[x])){
					boolean uTest = uri.startsWith("http://www.w3.org");
					if ((uTest == true) && (exc == false)){return true;}
					return false;
				}
				}
	    	}
	    return true;
	}

}
